import wx
from random import shuffle
import cPickle


###		###########		###
###		Preferences		###
###		###########		###

GAME_INFO = {	'in_game': False,
				'am_host': False, #DO: May be better way to do this?
				'server': ''}

RELEASE_INFO 		={	'VERSION' : 'Rummy.py v0.0.0.3 (Network friendly...)',
						'DATE' : 'September 10, 2007',
						'AUTHOR' : 'J. Kivinen',
						'CONTACT' : 'jl' + 'kiv' + 'ine' + '@' + 'lakehea' + 'du' + '.' + 'c' + 'a',
						'WEBSITE' : 'http://rummy-py.sourceforge.net\nhttp://www.jkiv.ca' }

DEFAULT_PREFERENCES	={	'min_players':2,
						'max_score':500,
						'network_name':'Player',
						'network_port':51015,
						'bot_skill':0,
						'image_path':'img/',
						'rule_aces_high':True,
						'rule_aces_low':True,
						'rule_discard_for_out':False,
						'rule_can_view_all_discarded':True}

PREFERENCES = DEFAULT_PREFERENCES	# Set default preferences to avoid un-set keys.  If un-set by load, it already is the default.

def LoadPreferences(preferences_file = '%srummy-py.preferences' % (PREFERENCES['image_path'].strip('img/')) ):
	"""Loads preferences from file. File should be formatted '<key:value>' with anything in between being ignored."""
	
	# DO: Is a pickle better?
	#		a) Pickle allows for faster loading
	#		b) If entire pickle fails, fallback on ALL defaults.
	#			(This method falls back on key-to-key default.)
	#		c) Already wrote this code.  However, the regular expression may be faulty.
	
	global PREFERENCES
	
	try:
		file = open(preferences_file, 'r')
		text = file.read()
		file.close()
		
		import re
		
		# Values of keys can contain anything but '<', '>', and ':' ([^:<>])
		# Though this might be a good policy to change it to \w or [A-Za-z0-9_ ]
		
		items = re.findall('<[^:<>]+:[^:<>]+>', text)
		# DON'T FORGET THIS EXPRESSION IS USED WHEN EDITING PREFERENCES (OnPref)
		
		for item in items:
			item.strip('<>')
			pair = item.split(':')
			if DEFAULT_PREFERENCES.has_key(pair[0]): # Avoid loading unused preferences.
				try:
					PREFERENCES[pair[0]] = int(pair[1]) # Integers
				except:
														# Booleans
					if pair[1].lowercase() == 'true': 
						PREFERENCES[pair[0]] = True			# T
					elif pair[1].lowercase() == 'false':
						PREFERENCES[pair[0]] = False		# F
					elif 'path' in pair[0] and pair[1][-1] != '/': 
						pair[1] += '/'						# Give paths trailing slashes
					else:								
						PREFERENCES[pair[0]] = pair[1]	# Strings
			
	except:
		PREFERENCES = DEFAULT_PREFERENCES
		
def SavePreferences(preferences_file = '%srummy-py.preferences' % (PREFERENCES['image_path'].strip('img/')) ):
	try:
		file = open(preferences_file, 'w')
		for key in PREFERENCES:
			file.write('<%s:%s>\n'%(key, PREFERENCES[key]))
		file.close()
	except:
		pass


###		##		###
###		AI		###
###		##		###


def Think():
	server = GAME_INFO['server']
	
	card_pickle = server.Pickup()
	card = cPickle.loads(card_pickle)[0]
	card_pickle = cPickle.dumps(card)
	server.Discard(card_pickle)
	server.EndTurn()

###		#####	###
###		Cards	###
###		#####	###

FACE_NAMES = [	'Ace',
				'Two',
				'Three',
				'Four',
				'Five',
				'Six',
				'Seven',
				'Eight',
				'Nine',
				'Ten',
				'Jack',
				'Queen',
				'King' ]
				
SUIT_NAMES = ['Spades', 'Clubs', 'Diamonds', 'Hearts']

class Card:
	def __init__(self,newSuit,newFace,newValue,newImg):
		self.suit = newSuit
		self.face = newFace
		self.value = newValue
		self.card_image = newImg
	
	def __str__(self):
		return "%s of %s" % (FACE_NAMES[self.face], SUIT_NAMES[self.suit],)

def Sort(pile):
		pile_length = len(pile)
		
		for face_value in range(pile_length):
			temp = pile[face_value]
			i = face_value
			while i > 0 and pile[i-1].face > temp.face:
				pile[i] = pile[i-1]
				i -= 1
			pile[i] = temp

		for suit_value in range(pile_length):
			temp = pile[suit_value]
			i = suit_value
			while i > 0 and pile[i-1].suit > temp.suit:
				pile[i] = pile[i-1]
				i -= 1
			pile[i] = temp

def MakeDeck(shuffle = False):
		
	deck = []
	for suit in range(0,4):
		for face in range(0,13):
			if face == 0:
				value = 20
			if face == 1:
				value = 40
			if face > 1 and face < 9:
				value = 5
			if face >= 9:
				value = 10
			image_source = "%d.gif" % ( (13*suit)+(face+1) ) # DO: Support non-*.gif
			deck.append( Card(suit, face, value, image_source) )

	if shuffle == True:
		shuffle(deck)
		
	return deck


def Deal(players, deck):
	number_of_players = len(players)
	
	cards_to_deal = 7
	
	'''
		Revise the following for network play:
	
		deal
			all - loop and deal 'player_id'th card to self, removing off-cards
			server begins tally of player card numbers
		all - pop card from pickup and place into
	'''
	
	for card_number in range(cards_to_deal):
		for player_id in range(number_of_players):
			card = deck.pop()
			players[player_id].hand.append(card)
			
			# DEBUG OUTPUTS
			print card, " given to Player ", player_id + 1
			# END DEBUG OUTPUTS

def PositionsToMeld(id, positions, list): #DO: Find out a more efficient way already in Python
	cards = []
	for position in positions:
		cards.append(list[position])
		
	return Meld(id, cards)


###		####	###
###		Meld	###
###		####	###

class Meld:
	def __init__(self, player_id, cards):
		self.cards = cards
		self.id = player_id
	
	def __str__(self):
		return "(%d) %s" % (self.id, self.cards)

def SplitMeld(meld):

	cards = []
	wildcards = []

	for card in meld.cards:
		if card.face == 1:
			wildcards.append(card)
		else:
			cards.append(card)	
	Sort(cards)

	return cards, wildcards

def IsSameSuitMeld(cards):

	is_same_suit_meld = True
	
	for card in cards:
		if card.suit != cards[0].suit:
			is_same_suit_meld = False	
			break

	return is_same_suit_meld

def IsMeld(meld): # Make sure 'cards' is a list of Card objects BEFORE passing to check
	
	cards, wildcards = SplitMeld(meld) 			# Break up cards in meld into ordered cards and wildcards.
			
	is_same_suit_meld = IsSameSuitMeld(cards) 	# Check assumes a certain type of meld.
		
	is_possible = True							# Establish that the meld is possible from the get-go.
		
	# DEBUG OUTPUTS
	print "Meld:"
	for card in cards:
		print card
	print "Wildcards:"
	for card in wildcards:
		print card
	print "Same Suit Meld?:", is_same_suit_meld
	# END DEBUG OUTPUTS
	
	'''
		(Make sure parameter 'cards' is a list of Card objects BEFORE passing to check.)
		
		Determines if Meld is valid.
			
			i.		Keeps track of # of wildcards to accept certain amount of fudge-factor in checking.
			ii.		Determine type of meld is it? See IsSameSuitMeld().
			iii.	Check first card and second card.
			iv.		Do standard meld check (is it same suit and one value away and etc...)
			v.		If there's a gap (ie. 3 and 6) try to fill with wildcards.  Fail if cannot.
			vi.		Do special case where A is high.[1]
			vii.		Do special case where 3 or 4 two's.
			viii.	Otherwise, not possible
			
					[1] Lowest_Card_Face - (Highest_Possible_Value - Total_Number_of_cards_in_Meld) cannot be < 2 for Ace to be high.
	'''

		
	wildcards_left = len(wildcards)			
	
	if len(cards) > 0:

		card_a = cards[0]
			
		for i in range(len(cards)-1):
		
			card_b = cards[i+1]
				
			if (is_same_suit_meld == False and card_a.face == card_b.face ) or \
				(is_same_suit_meld == True and ( \
					(card_a.face == card_b.face - 1) or \
					(card_a.face == 0 and card_b == cards[1] and ( card_b.face - ( 12 - ( len(cards) + len(wildcards) ) ) ) >= 2 )  \
					) \
				):
				#
				# Removed the following condition after first "is_same_suit_meld == True and..."
				#	due to possible uselessness.
				#
				# 		(card_a.face == (12) and cards[0].face == 0) or \
				
				card_a = card_b
				
			elif len(wildcards) >= (card_a.face - card_b.face - 1):
				for space in range(card_a.face - card_b.face - 1):
					wildcard = wildcards.pop()
					cards.insert(i + 1 + space, wildcard)
			else:
				is_possible = False
				break
				
	elif wildcards_left >= 3 and len(cards) == 0: # All wildcard meld
		is_possible = True
	else:
		# The magic case where 3 cards or more come in, less than that are wildcards, and none are not wildcards.
		is_possible = False
		
	# DEBUG OUTPUTS
	print "Resulting meld: "
	for card in cards:
		print card
	print "Is possible: ", is_possible
	# END DEBUT OUTPUTS

	if len(wildcards) > 0:
		# DO: Ask for placement of extra wildcards (before/after)
		cards.extend(wildcards)

	#
	#	DO: Check wildcards for conflicts if already in play.
	#
	
	# Does meld return?
	return is_possible

###		##########		###
###		Networking		###
###		##########		###

def StartNetwork():
	from SimpleXMLRPCServer import SimpleXMLRPCServer

	server = SimpleXMLRPCServer(('0.0.0.0', PREFERENCES['network_port']))
	
	### Registering functions
	server.register_function(Initialize, 'Initialize')
	server.register_function(Pickup, 'Pickup')
	server.register_function(Discard, 'Discard')
	server.register_function(Melder, 'Meld') # Distinguish class and method...
	server.register_function(Sync, 'Sync')
	server.register_function(GetTurnFlags, 'GetTurnFlags')
	server.register_function(CheckStatus, 'CheckStatus')
	server.register_function(EndTurn, 'EndTurn')
	### End registering functions
	
	server.serve_forever()

# Client-side

def JoinGame(ip, port):
	'''Called to attempt to join a networked game.  Returns the local player's id, rules, and hand.'''

	import xmlrpclib
	
	uri = 'http://' + str(ip) + ':' + str(port) + '/'
	
	server = xmlrpclib.Server(uri)
	
	global GAME_INFO
	
	GAME_INFO['in_game'] = True
	GAME_INFO['am_host'] = False
	GAME_INFO['server'] = server
	
	local_player_pickle, meld_pile_pickle, discard_pile_pickle = server.Initialize(PREFERENCES['network_name'])
	
	global local_player
	global meld_pile
	global discard_pile

	# Unpickle		
	local_player = cPickle.loads(local_player_pickle)
	meld_pile = cPickle.loads(meld_pile_pickle)
	discard_pile = cPickle.loads(discard_pile_pickle)

	print "You joined game at %s as %s (id: %d)" % (uri, local_player.name, local_player.id)

# Server-side

def Initialize(name):
	new_player = None
	for player in players:
		if player.is_bot == True:
			player.is_bot = False
			player.name = name
			new_player = player
			break

	print "%s (%d) joined game." % (new_player.name, new_player.id)
	
	new_player_pickle = cPickle.dumps(new_player)
	meld_pile_pickle = cPickle.dumps(meld_pile)
	discard_pile_pickle = cPickle.dumps([discard_pile])
	
	return new_player_pickle, meld_pile_pickle, discard_pile_pickle


def Pickup(discard = False, all = False):

	global current_player

	if discard == True and all == True:
		cards = meld_pile
		cards_pickle = cPickle.dumps(cards)
	
	elif discard == True and all == False:
		cards = [discard_pile.pop()]
		
	else:
		cards = [pickup_pile.pop()]
	
	cards_pickle = cPickle.dumps(cards)

	current_player.turn_flags = [True, False]
	
	return cards_pickle
	
def Melder(meld_pickle):

	global meld_pile

	meld = cPickle.loads(meld_pickle)
	
	if IsMeld(meld) == True:
	
		for player in players:
			if player.id == meld.id:
				for card in meld:
					player.score += card.value
				break
		meld_pile.append(meld)
		
		return True
	else:
		return False	
		
def Discard(card_pickle):
	global discard_pile
	
	card = cPickle.loads(card_pickle)
	
	discard_pile.append(card)
	
	current_player.turn_flags = [True, True]
	
	return ''

def EndTurn():

	global current_player
	global discard_pile
	global pickup_pile

	# Replace empty pickup pile if empty
	if len(pickup_pile) == 0:
		print "Empty pickup pile..." # DEBUG OUTPUT
		top_card = discard_pile.pop()
		shuffle(discard_pile)
		for card in discard_pile: 
			pickup_pile.append(card)
		discard_pile = []
		discard_pile.append(top_card)

	if current_player.id == len(players) - 1:
		current_player = players[0]
	else:
		current_player = players[current_player.id + 1]
	
	current_player.turn_flags = [False, False]
	
	if current_player.is_bot == True:
		
		###	Start AI thread
		from threading import Thread
		
		ai_thread = Thread(target=Think)
		ai_thread.start()
		###

	return '' # DO: Find out what allow_none is.
	
def GetTurnFlags(id):
	return players[id].turn_flags

def CheckStatus(id):

	if current_player.id == id:
		turn_bool = True
	else:
		turn_bool = False
	
	return turn_bool, current_player.id

def Sync():
	meld_pile_pickle = cPickle.dumps(meld_pile) 
	discard_pile_pickle = cPickle.dumps(discard_pile) # Return as list for ReDraw()
	
	return meld_pile_pickle, discard_pile_pickle

###		#######		###
###		Players		###
###		#######		###

class Player:
	def __init__(self, id = -1, ip = 'localhost', is_bot = True):
		self.hand = []
		self.score = 0
		self.id = id
		self.ip = ip # DEFUNCT?
		self.is_bot = is_bot
		self.name = ''
		self.turn_flags = [True, True] # (picked_up, discarded)
		self.last_check = None
		
def MakePlayersList(number_of_players):

	players_list = []

	for i in range(number_of_players):
		players_list.append(Player(id = i))
		
	return players_list


###		###		###
###		GUI		###
###		###		###

class MainWindow (wx.Frame):
	def __init__(self, parent, ID, title):
		wx.Frame.__init__(self, parent, -1, title, wx.DefaultPosition, wx.Size(800, 600), style = wx.DEFAULT_FRAME_STYLE | wx.FULL_REPAINT_ON_RESIZE)

		menubar = wx.MenuBar()
		
		file = wx.Menu()
		game = wx.Menu()
		help = wx.Menu()
		
		file.Append(101, '&New Game', 'Start a new game.')
		file.Append(102, '&Join Game', 'Join a network game.')
		file.AppendSeparator()
		file.Append(103, '&Preferences', 'Edit your game preferences.')
		file.AppendSeparator()
		file.Append(110, '&Quit', 'Quit?')
		
		game.Append(201, 'Sort hand', 'Sort hand to see possible runs.')
		
		help.Append(301, '&About', 'About this program.')
		help.Append(302, '&Help', "Don\'t know how to play?")
		menubar.Append(file, '&File')
		menubar.Append(game, '&Game')
		menubar.Append(help, '&Help')
		self.SetMenuBar(menubar)
		self.CreateStatusBar()
		
		self.Bind(wx.EVT_MENU, self.OnNew, id=101)
		self.Bind(wx.EVT_MENU, self.OnJoin, id=102)
		self.Bind(wx.EVT_MENU, self.OnPref, id=103)
		self.Bind(wx.EVT_MENU, self.OnQuit, id=110)

		self.Bind(wx.EVT_MENU, self.OnSortHand, id=201)
				
		self.Bind(wx.EVT_MENU, self.OnAbout, id=301)
		self.Bind(wx.EVT_MENU, self.OnHelp, id=302)
		
		self.selected_cards = []
		self.hand_buttons = []
		self.meld_buttons = []
		self.discard_button = None
		self.other_buttons = []

	#File
	def OnNew(self, event):
		
		global GAME_INFO
		
		if GAME_INFO['in_game'] == True:
			GAME_INFO['server'].LeaveGame()

		global players
		global pickup_pile
		global local_player
		
		players = MakePlayersList(PREFERENCES['min_players'])

		local_player = players[0]
		pickup_pile = MakeDeck()
		
		local_player.is_bot = False
		local_player.turn_flags = [False, False]
		
		###	Start server thread
		from threading import Thread
		
		server_thread = Thread(target=StartNetwork)
		server_thread.start() # DO: Kill this later after completely finished game.
		###
		
		import xmlrpclib
		local_client = xmlrpclib.Server('http://localhost:%d' % (PREFERENCES['network_port']))
		
		GAME_INFO['am_host'] = True
		GAME_INFO['server'] = local_client
		
		self.NewRound()
		self.ReDraw()
		
	def OnJoin(self, event):
		join_dialog = wx.TextEntryDialog(self, 'Enter IP: ','Join Game')
		if join_dialog.ShowModal() == wx.ID_OK:
			connect_to = join_dialog.GetValue()
			JoinGame(connect_to, PREFERENCES['network_port'])
		join_dialog.Destroy()

		self.Sync()
		self.ReDraw()
		self.WaitTurn()
		
	def OnPref(self, event):
		preferences_dialog = self.PreferencesDialog(self, -1, 'Preferences')
		if preferences_dialog.ShowModal() == wx.ID_OK:
			
			PREFERENCES['network_name'] = preferences_dialog.txt_name.GetValue()
			PREFERENCES['min_players'] = preferences_dialog.sc_min_players.GetValue()
			PREFERENCES['max_score'] = preferences_dialog.sc_max_score.GetValue()
			
			SavePreferences()
			# PREFERENCES = LoadPreferences()
			
		preferences_dialog.Destroy()

	def OnQuit(self, event):
		GAME_INFO['server'].LeaveGame()
		self.Close()
		
	#About
	def OnAbout(self, event):
		about = "%s \n%s \n\nAuthor: %s \n%s \n\nWebsite: \n%s" % (RELEASE_INFO['VERSION'], RELEASE_INFO['DATE'], RELEASE_INFO['AUTHOR'], RELEASE_INFO['CONTACT'], RELEASE_INFO['WEBSITE'],)
		about_dialog = wx.MessageDialog(self, about, 'About Rummy.py', wx.OK|wx.ICON_INFORMATION)
		about_dialog.ShowModal()
		about_dialog.Destroy()
	
	def OnHelp(self, event):
		print "No help for you! (Right now.)" #DO: Offer help.
	
	
	### Drawing Methods
	
	def ReDrawHand(self):
		(maxx, maxy) = self.GetClientSizeTuple()
	
		for button in self.hand_buttons:
			button.Destroy()
	
		self.hand_buttons = []
		
		hand = local_player.hand
		hand_length = len(hand)
		
		for i in range(hand_length):
			bmp = wx.Image('%s%s'%(PREFERENCES['image_path'],hand[i].card_image),wx.BITMAP_TYPE_GIF).ConvertToBitmap() # DO: suppport for non-*.gif
			if i in self.selected_cards:
				card_button = wx.BitmapButton(self, i, bmp, \
					pos=( i * maxx/(6 + hand_length) + maxx/2 - (maxx * hand_length) / (2 * (6 + hand_length)), \
						maxy - 160), \
						style = wx.NO_BORDER) # DO: Debug wx.NO_BORDER on OS X
	
			else:
				card_button = wx.BitmapButton(self, i, bmp, \
					pos=( i * maxx/(6 + hand_length) + maxx/2 - (maxx * hand_length) / (2 * (6 + hand_length)), \
					maxy - 120), \
					style = wx.NO_BORDER)
					# Horizontal position:
					# relative position of current card + (middle - half the length of the final drawing to center justify)
					# Vertical position: 120px from bottom
	
			card_button.Bind(wx.EVT_BUTTON, self.OnCard)
			self.hand_buttons.append(card_button)
			
	def ReDrawMelds(self):
		# Draw melds
		if len(meld_pile) > 0:
			(maxx, maxy) = self.GetClientSizeTuple()
		
			for button in self.meld_buttons:
				button.Destroy()

			self.meld_buttons = []

			for meld_number, meld in enumerate(meld_pile):
				vertical_space = 30*meld_number

				for card_number, card in enumerate(meld.cards):
					if meld_pile[meld][card_number].face == 1:
						bmp = wx.Image('%s%s'%(PREFERENCES['image_path'], card.card_image),wx.BITMAP_TYPE_GIF).ConvertToBitmap()
						card_button = wx.BitmapButton(self, 3000+(50*meld_number)+card_number, bmp, pos=( maxx - 80 - 16*len(meld.cards) + 16*card_number, vertical_space), style = wx.NO_BORDER)
						card_button.Bind(wx.EVT_BUTTON, self.ReplaceTwo)
						self.meld_buttons.append(card_button)
					else:
						bmp = wx.Image('%s%s'%(PREFERENCES['image_path'], card.card_image),wx.BITMAP_TYPE_GIF).ConvertToBitmap()
						card_button = wx.BitmapButton(self, 3000+(50*meld_number)+card_number, bmp, pos=( maxx - 80 - 16*len(meld.cards) + 16*card_number, vertical_space), style = wx.NO_BORDER)
						card_button.Bind(wx.EVT_BUTTON, self.ExtendMeld)
						self.meld_buttons.append(card_button)
						
	def ReDrawDiscard(self):
		(maxx, maxy) = self.GetClientSizeTuple()
	
		if self.discard_button:
			self.discard_button.Destroy()

		self.discard_button = None

		if len(discard_pile) == 0: # Blank discard pile
			bmp = wx.Image('%sx.png'%(PREFERENCES['image_path']),wx.BITMAP_TYPE_PNG).ConvertToBitmap()
			card_button = wx.BitmapButton(self, 2000, bmp, pos=(maxx/2, maxy/2 - 80), style = wx.NO_BORDER)
			card_button.Bind(wx.EVT_BUTTON, self.OnDiscard)
		else:
			bmp = wx.Image('%s%s'%(PREFERENCES['image_path'], discard_pile[-1].card_image),wx.BITMAP_TYPE_GIF).ConvertToBitmap()
			card_button = wx.BitmapButton(self, 2000, bmp, pos=(maxx/2, maxy/2 - 80), style = wx.NO_BORDER)
			card_button.Bind(wx.EVT_BUTTON, self.OnDiscard)
		
		self.discard_button = card_button
			
	def ReDraw(self):
		(maxx, maxy) = self.GetClientSizeTuple()
		
		for button in self.other_buttons:
			button.Destroy()
		
		self.other_buttons = []
		
		# Draw pickup pile
		bmp = wx.Image('%sb.png'%(PREFERENCES['image_path']),wx.BITMAP_TYPE_PNG).ConvertToBitmap()
		card_button = wx.BitmapButton(self, 2001, bmp, pos=(3*maxx/8, maxy/2 - 80), style = wx.NO_BORDER)
		card_button.Bind(wx.EVT_BUTTON, self.OnPickup)
		self.other_buttons.append(card_button)		

		# Draw 'View Pile' button
		card_button = wx.Button(self, 2002, 'View Pile', pos=(maxx/2-1, maxy/2 + 30))
		card_button.Bind(wx.EVT_BUTTON, self.ViewDiscardPile)
		self.other_buttons.append(card_button)
		
		if len(discard_pile) > 1 and PREFERENCES['rule_can_view_all_discarded'] == True:
			card_button.Enable(True)
		else:
			card_button.Enable(False)
		
		# Draw 'Meld' button
		card_button = wx.Button(self, 2003, 'Meld', pos=(maxx/2-1, maxy/2 + 65))
		card_button.Bind(wx.EVT_BUTTON, self.PlayMeld)
		self.other_buttons.append(card_button)
		
		self.ReDrawHand()
		self.ReDrawMelds()
		self.ReDrawDiscard()
	
	### End Drawing Methods
	
	def OnSortHand(self, event):
		Sort(local_player.hand)
		self.ReDrawHand()
		
	def OnCard(self, event):
		id = event.GetId()
	
		if GAME_INFO['server'].GetTurnFlags(local_player.id) == [True, False]:
			if id in self.selected_cards:
				self.selected_cards.remove(id)
			else:
				self.selected_cards.append(id)
			
			self.ReDrawHand()
		
	def OnPickup(self, event):
			
		if GAME_INFO['server'].GetTurnFlags(local_player.id) == [False, False]:
			card_pickle = GAME_INFO['server'].Pickup()
			card = cPickle.loads(card_pickle)
			local_player.hand.extend(card)

			self.ReDrawHand()
		
	def OnDiscard(self, event):
	
		# Pickup the top discard pile card
		if GAME_INFO['server'].GetTurnFlags(local_player.id) == [False, False]:
			card_pickle = GAME_INFO['server'].Pickup(True)
			card = cPickle.loads(card_pickle)
			
			local_player.hand.extend(card)
			
			self.Sync()
			self.ReDrawHand()
			self.ReDrawDiscard()
		
		# Discard
		elif len(self.selected_cards) == 1 and GAME_INFO['server'].GetTurnFlags(local_player.id) == [True, False]:
		
			discard_card = local_player.hand[self.selected_cards[0]]
			del local_player.hand[self.selected_cards[0]]
			del self.selected_cards[0]
			
			discard_card_pickle = cPickle.dumps(discard_card)
			GAME_INFO['server'].Discard(discard_card_pickle)

			if GAME_INFO['am_host'] == False:
				self.Sync()

			self.ReDrawHand()
			self.ReDrawDiscard()
			
			if len(local_player.hand) == 0:
				GAME_INFO['server'].EndRound()
				
				# DO: Networked End-round...
				#		perhaps open dialog and get a scoresheet
				#		then draw the scoresheet.
				#		(in EndOfRoundDialog)
				
				end_of_round_dialog = EndOfRoundDialog(self, -1, 'Rummy!')
				end_of_round_dialog.ShowModal()
				end_of_round_dialog.Destroy()
			else:
				GAME_INFO['server'].EndTurn()
				self.WaitTurn()
			
	def PlayMeld(self, event):
		if GAME_INFO['server'].GetTurnFlags(local_player.id) == [True, False] and len(self.selected_cards) >= 3:

			global local_player

			meld = PositionsToMeld (local_player.id, self.selected_cards, local_player.hand)
			meld_pickle = cPickle.dumps(meld)
			
			if GAME_INFO['server'].Meld(meld_pickle) == [True]:
				for position in self.selected_cards:
					del local_player.hand[position]
				self.Sync()
			else:
				print "Not a meld." # DEBUG OUTPUT
				
			self.ReDrawMelds()
					
	def ReplaceTwo(self, event):
		if GAME_INFO['server'].GetTurnFlags(local_player.id) == [True, False] and len(self.selected_cards) == 1:
			# Reminder: make new list, meld_pileOwners for player IDs who made the meld
			#	or... give player objects 'melds' with IDs of melds.
			print "Swapping not yet implemented." # DEBUG OUTPUT
			
			# take that meld,
			# remove 2
			# add cards
			# run meld check
			# if return True
			#	remove score of the 2 from owner's score
			#	add score of the replacement card to owner's score
			#	swap player's card with 2

	def ExtendMeld(self, event):
		if GAME_INFO['server'].GetTurnFlags(local_player.id) == [True, False] and len(self.selected_cards) > 0:
			print "Extending melds not yet implemented." # DEBUG OUTPUT
			# take the meld
			# add cards
			# run meld check
			# if return True
			#	give player points for 

	def ViewDiscardPile(self, event):
		view_discard_pile_dialog = self.ViewDiscardPileDialog(self, -1, 'Discard pile')
		view_discard_pile_dialog.ShowModal()
		view_discard_pile_dialog.Destroy()

	def WaitLoop(self):
		from time import sleep
		
		status = GAME_INFO['server'].CheckStatus(local_player.id) # returns [local turn bool, current player id]
		
		while True:
			sleep(1)	# Space out check as to not crash server.
			old_status = status
			status = GAME_INFO['server'].CheckStatus(local_player.id)
			if status[1] != old_status[1]:	# Update screen on new turn
				self.Sync()
				wx.CallAfter(self.ReDrawDiscard) # For threaded screen-redraw... seems to work.
				# wx.CallAfter(self.ReDrawMelds)
				
			if status[0] == True:
				break
		
	def WaitTurn(self):
		# DO: URGENT REDRAW SCREEN W/ THREAD BREAKS
	
		###	Start WaitLoop thread
		from threading import Thread
		
		server_thread = Thread(target=self.WaitLoop)
		server_thread.start()
		###
	
	def Sync(self):
		
		meld_pile_pickle, discard_pile_pickle = GAME_INFO['server'].Sync()
		
		global discard_pile
		global meld_pile
		
		meld_pile = cPickle.loads(meld_pile_pickle)
		discard_pile = cPickle.loads(discard_pile_pickle)

	def NewRound(self):
		global pickup_pile
		global current_player
		
		current_player = players[0] # DO: keep round numbers to rotate starting player
		
		shuffle(pickup_pile)
		
		Deal(players, pickup_pile)
		discard_pile.append(pickup_pile.pop())
		

	def EndRound(self):
		# DEBUG OUTPUTS
		print "*** Round finished ***"
		# END DEBUG OUTPUTS
	
		# DO: Edit this for network play?
		#		Keep track locally?
		#		Have score returned?
	
		for player_id in range (len(players)):
			for card in players[player_id]:
				card = players[player_id].hand.pop() # DO: Avoid pop.
				players[player_id].addScore(-1 * card.value())
				pickup_pile.append(card)
			
			# DEBUG OUTPUTS
			print "Player ", player_id + 1, " has a score of ", players[player_id].score
			# END DEBUG OUTPUTS
			
		pickup_pile.extend(discard_pile)
		del discard_pile [:]
	
		for meld in range(len(meld_pile)):
			pickup_pile.extend(meld_pile[meld])
		del meld_pile [:]
	
		# DEBUG OUTPUTS
		print len(meld_pile)
		print len(discard_pile)
		print len(pickup_pile)
		# END DEBUG OUTPUTS


	### Dialog Classes	###
	
	class PreferencesDialog (wx.Dialog):
		def __init__(self, parent, ID, title):
			wx.Dialog.__init__(self, parent, -1, title, wx.DefaultPosition, wx.Size(400, 180))
	
			mainBox = wx.BoxSizer(wx.VERTICAL)
			hBox1 = wx.BoxSizer(wx.HORIZONTAL)
			hBox2 = wx.BoxSizer(wx.HORIZONTAL)
			hBox3 = wx.BoxSizer(wx.HORIZONTAL)
			hBox4 = wx.BoxSizer(wx.HORIZONTAL)
			hBox5 = wx.BoxSizer(wx.HORIZONTAL)
			
			preferred_name = PREFERENCES['network_name']
			preferred_min_players = PREFERENCES['min_players']
			preferred_max_score = PREFERENCES['max_score']
			
			label_name = wx.StaticText(self, -1, 'Name: ')
			self.txt_name = wx.TextCtrl(self, 1, value = preferred_name)
			
			label_min_players = wx.StaticText(self, -1, 'Number of Players: ')
			self.sc_min_players = wx.SpinCtrl(self, -1, style = wx.SP_WRAP, min = 2, max = 6, initial = preferred_min_players)
			
			label_max_score = wx.StaticText(self, -1, 'Maximum Score: ')
			self.sc_max_score = wx.SpinCtrl(self, -1, style = wx.SP_WRAP, min = 0, max = 1000000, initial = preferred_max_score)
			
			# When in a game, disable changing your preferences.
			#	Might even be a better idea to switch game-specific
			#	values to it's own dictionary.
			if GAME_INFO['in_game'] == True:
				self.sc_max_score.Enable(False)
				self.sc_min_players.Enable(False)
			
			buttons =  self.CreateButtonSizer(wx.CANCEL|wx.OK)
	
			hBox1.Add((20,20),0,wx.EXPAND)
			hBox1.Add(label_name,10,wx.ALIGN_TOP)
			hBox1.Add(self.txt_name,10,wx.ALIGN_TOP)
			hBox1.Add((20,20),0,wx.EXPAND)
			
			hBox2.Add((20,20),0,wx.EXPAND)
			hBox2.Add(label_min_players,20,wx.ALIGN_TOP)
			hBox2.Add(self.sc_min_players,10,wx.ALIGN_TOP)
			hBox2.Add((20,20),0,wx.EXPAND)
	
			hBox3.Add((20,20),0,wx.EXPAND)
			hBox3.Add(label_max_score,20,wx.ALIGN_TOP)
			hBox3.Add(self.sc_max_score,10,wx.ALIGN_TOP)
			hBox3.Add((20,20),0,wx.EXPAND)
	
			hBox4.Add((20,20),0,wx.EXPAND)
			hBox4.Add(wx.StaticLine(self, -1),30,wx.EXPAND)
			hBox4.Add((20,20),0,wx.EXPAND)
			
			hBox5.Add(buttons,1,wx.ALIGN_RIGHT)
	
			mainBox.Add((20,20),0,wx.EXPAND)
			mainBox.Add(hBox1,10,wx.EXPAND)
			mainBox.Add(hBox2,10,wx.EXPAND)
			mainBox.Add(hBox3,10,wx.EXPAND)
			mainBox.Add(hBox4,10,wx.EXPAND)
			mainBox.Add(hBox5,10,wx.EXPAND)
			
			self.SetSizer(mainBox)
			self.SetAutoLayout(1)
			
			
	class ViewDiscardPileDialog (wx.Dialog):
		def __init__(self, parent, ID, title):
		
			window_width = len(discard_pile)*16 + 200
			window_height = 200
			
			wx.Dialog.__init__(self, self.parent, -1, title, wx.DefaultPosition, wx.Size(window_width, window_height))
			
			x = 16 # Might need to change factor
			for card in range(len(self.parent.discard_pile)):
				bmp = wx.Image(self.parent.discard_pile[card].card_image,wx.BITMAP_TYPE_GIF).ConvertToBitmap()
				wx.StaticBitmap(self, -1, bmp, pos=(( 16 * len(self.parent.discard_pile) - x ), window_height/2 - 48))
				x += 16
			view_discard_option = wx.Button(self, 2101, 'Take pile', pos=(window_width - 100, 40))
			view_discard_option.Bind(wx.EVT_BUTTON, self.OnTakeAll, id = 2101)
			view_discard_option = wx.Button(self, 2102, 'Leave pile', pos=(window_width - 100, 70))
			view_discard_option.Bind(wx.EVT_BUTTON, self.OnCancel, id = 2102)
	
		def OnCancel(self, event):
			self.Close()
	
		def OnTakeAll(self, event):
			global local_player
			global discard_pile
			
			if GAME_INFO['server'].GetTurnFlags(local_player.id) == [False, False]:
				cards_pickle = GAME_INFO['server'].Pickup(all = True)
				cards = cPickle.loads(cards_pickle)
				local_player.hand.extend(cards)
				discard_pile = []
				
			self.Close()
				
			
	class EndOfRoundDialog (wx.Dialog):
		def __init__(self, parent, ID, title):		
			wx.Dialog.__init__(self, parent, -1, title, wx.DefaultPosition, wx.Size(200, 500))
			
			self.parent = parent
			
			# DO: Insert score table here.
			# TEMP
			for player in self.parent.players:
				print player.name, player.id, player.score # player has [name, player_id, score]
			# END TEMP
			
			continue_button = wx.Button(self, 3001, 'Next Round', pos=(0,0))
			continue_button.Bind(wx.EVT_BUTTON, self.OnContinue)
	
			for player in self.parent.players:
				if player.score >= PREFERENCES['max_score']:
					continueButton.Enable(False)
					# Add other conditions for win...
					break
				
		def OnContinue(self, event):
			self.parent.NewRound()	
			self.Close()

class Rummy(wx.App):
	def OnInit(self):
		frame = MainWindow(None, -1, "%s" % (RELEASE_INFO['VERSION'],) )
		frame.Show(True)
		return True


	
### SET VARIABLES ###

LoadPreferences()

local_player = Player()
current_player = None

players = []

discard_pile = [] # Server has all, client has 1 (top)
meld_pile = [] # Server has all, client has all
pickup_pile = [] # Server has all, client has none.

### END SET VARIABLES ###


#======PROGRAM STARTS HERE========

app = Rummy(0)
app.MainLoop()
